// MCU : PIC 16F887
// External clock : Quartz 4 MHz
// LED YELLOW is connected to pin 17 (portc.b2).
// LM35 temperature sensor : pin GND to GND,
//                           pin DQ to pin 5 (AN3),
//                           pin VCC to VCC.

/******************************************************************************/

// LCD module connections :
                                        // VSS (pin  1) -> GND
                                        // VDD (pin  2) -> VCC
                                        // VEE (pin  3) -> middle pin contrast potentiometer
sbit LCD_RS at RB4_bit;                 // RS  (pin  4) -> portB.b4 (pin 37)
// RW not used.                         // RW  (pin  5) -> GND
sbit LCD_EN at RB5_bit;                 // EN  (pin  6) -> portB.b5 (pin 38)
                                        // D0  (pin  7) -> GND
                                        // D1  (pin  8) -> GND
                                        // D2  (pin  9) -> GND
                                        // D3  (pin 10) -> GND
sbit LCD_D4 at RB0_bit;                 // D4  (pin 11) -> portB.b0 (pin 33)
sbit LCD_D5 at RB1_bit;                 // D5  (pin 12) -> portB.b1 (pin 34)
sbit LCD_D6 at RB2_bit;                 // D6  (pin 13) -> portB.b2 (pin 35)
sbit LCD_D7 at RB3_bit;                 // D7  (pin 14) -> portB.b3 (pin 36)
                                        //     (pin 15) -> GND
                                        //     (pin 16) -> GND

sbit LCD_RS_Direction at TRISB4_bit;
sbit LCD_EN_Direction at TRISB5_bit;
sbit LCD_D4_Direction at TRISB0_bit;
sbit LCD_D5_Direction at TRISB1_bit;
sbit LCD_D6_Direction at TRISB2_bit;
sbit LCD_D7_Direction at TRISB3_bit;

/******************************************************************************/

#define CMD 0
#define DATA 1
#define LCD_PORT PORTB
#define LCD_COLUMNS 16

/******************************************************************************/

// Declaration of variables :

char texte[64];                         // 0 to 255
                                        // Variables ci-dessous sorties du main() pour les dclarer en global.
                                        // Le PIC 16F a moins de mmoire RAM et ROM que le PIC 18F.
unsigned char *pointeur_de_char;        // Dclaration d'un pointeur (*) de char "pointeur_de_char".

unsigned char i;

char ROW[] = {0x80, 0xC0};

/******************************************************************************/

// Routines :

void ADC_initialization()
{
 ADCON0 = 0b01000001;              // b0 = 1 (ADON) : le convertisseur A/N interne est activ.
                                   // b1 (GO/DONE)  : A/D Conversion Status bit.
                                   // b2, b3, b4, b5 (CHS0, CHS1, CHS2, CHS3) : Analog Channel Select bits.
                                   // b6 = 1 (ADSC0), b7 = 0 (ADSC1)          : fosc / 8 = 4 MHz / 8 = 0,5 MHz = 500 KHz
                                   //                                           tosc = 1 / fosc = 1 / 500 KHz = 0,002 mS = 2 uS.
                                   // 2 uS est le temps de conversion A/D d'un bit, TAD.
                                   // Pour une conversion totale sur 10 bits il faut 12 TAD.
                                   // Pour une conversion correcte il faut que TAD = 1,6 uS au minimum.
                                   // TAD = 2 uS / bit. 12 TAD = 24 uS pour 10 bits.

 ADCON1 = 0b10000000;              // b0 = b1 = b2 = b3 = 0 : Unimplemented. Read as '0'.
                                   // b4 = 0 (VCFG0)        : Voltage reference VDD.
                                   // b5 = 0 (VCFG1)        : Voltage reference VSS.
                                   // b6 = 0                : Unimplemented. Read as '0'.
                                   // b7 = 1                : ADFM (A/D result ForMat) = 1 (format justifi  droite).
                                   // Explication :
                                   //          ADRESH        |         ADRESL
                                   // b7|b6|b5|b4|b3|b2|b1|b0|b7|b6|b5|b4|b3|b2|b1|b0
                                   //  0| 0| 0| 0| 0| 0| r| r| r| r| r| r| r| r| r| r
                                   //    les 6 MSB = 0 |   rsultat r sur 10 bits
}

unsigned int ADC_read(unsigned char channel)
// Si message d'erreur "'ADC_read' Identifier redefined" pendant compilation, dans Library Manager dcocher ADC_Read.
// ADC_read est ma propre routine crite pour apprentissage. Je n'utilise pas ADC_Read de la librairie Mikroelektronika.
{
 static unsigned int k;                                           // Variable locale (static).
 ADCON0 = ADCON0 & 0b11000011;                                    // ADCON0 = 01000001
                                                                  //        & 11000011 (masque)
                                                                  // ADCON0 = 01000001
                                                                  // masque :
                                                                  // b0 = 1 (ADON)                               : pour prendre en compte l'tat du convertisseur A/D, en service (0) ou  l'arrt (1).
                                                                  // b1 = 1 (GO/DONE)                            : pour prendre en compte l'tat de la conversion A/D, en cours (1) ou termine (0).
                                                                  // b2 = b3 = b4 = b5 = 0 (CHS0 CHS1 CHS2 CHS3) : pour re initialiser au premier canal de conversion, le canal 0 (CHS0 = CHS1 = CHS2 = CHS3 = 0).
                                                                  // b6 = b7 = 1 (ADSC0 ADSC1)                   : pour prendre en compte la valeur du diviseur, de la vitesse de conversion, choisi.
 channel = channel << 2;                                          // Dcalage channel de 2 bits  gauche pour placer la valeur de channel dans bits b2 b3 b4 b5 (CHS0 CHS1 CHS2 CHS3).
 ADCON0 = ADCON0 | channel;                                       // OU logique bit  bit entre ADCON0 et channel.
 // ou ADCON0 |= channel;
 Delay_ms(2);                                                     // Dlais de 2 mS minimum. >>> A ajuster si besoin. <<<
 ADCON0.GO_DONE = 1;                                              // Dclenchement de la conversion A/N.
                                                                  // ou ADCON0.b2 = 1;
 _asm NOP;                                                        // Recommand par Microchip.
 while (ADCON0.GO_DONE == 1);                                     // Attendre que le bit GO.DONE passe  0.
 {}
 k = (unsigned int)ADRESL + (unsigned int)(ADRESH * 256);         // XXX A REVOIR XXX
 //ou k = (unsigned int)ADRESL + (unsigned int)(ADRESH << 8);
 return(k);
}

void LCD_E_Pulse(void)
{
 LCD_EN = 1;
 Delay_us(8);                                                     // Delay between 5 uS and 10 uS.
 LCD_EN = 0;
 Delay_us(500);                                                   // Delay between 5 uS and 1000 uS (1 mS).
}

void LCD_Write(unsigned char cmd_or_data, unsigned char byte)
{
 unsigned char low, high;
 high = (byte >> 4) & 0x0F;                                       // b7 b6 b5 b4 of byte are shifted to b3 b2 b1 b0 of byte and saved in high.
 low = byte & 0x0F;                                               // b3 b2 b1 b0 of byte are saved in low.
 LCD_PORT = high;                                                 // Send higher nibble of byte to the LCD.
 LCD_RS = cmd_or_data;
 LCD_E_Pulse();
 LCD_PORT = low;                                                  // Send lower nibble of byte to the LCD.
 LCD_RS = cmd_or_data;
 LCD_E_Pulse();
}

void StrConstRamCpy(unsigned char *dest, const code char *source) // Copie le texte de la FLASH ROM vers la RAM.
{
 while (*source)*dest ++ = *source ++;
 *dest = 0;                                                       // Terminateur "0" fin de chaine de caractre.
}

void LCD_Write_String(char *msg)                                  // Variante avec pointeur msg non modifi.
{
 int k;
 k = 0;
 while(*(msg + k) > 0)
 {
  LCD_Write(DATA, (*(msg + k)));                                  // Data pointe par (msg + k).
  k++;
  if (k == LCD_COLUMNS) break;                                    // Si k = 16 sortie de la boucle while ...
 }
}

void LCD_Write_String_At(char line, char column, char *msg)
{
 LCD_Write(CMD, ROW[line]|(column & 0x0F));                       // Print message on desired line and desired column.
 LCD_Write_String(msg);                                           // OK.
}

void float2ascii_v1 (float x, unsigned char *str, char precision) // Converts a floating point number to an ASCII string.
{                                                                 // Version limited at 5 decimals maximum.
 int ie, i, k, ndig;                                              // x is stored into str, which should be at least 30 chars long.
 double y;

 if (precision >= 5)
    {
     precision = 5;
    }
    else
       {
        precision++;
       }

 ndig = precision;

 ie = 0;

 if (x < 0.00000)                                                 // If x is negative, write minus and reverse.
    {
     *str++ = '-';
     x = -x;
    }

 if (x > 0.000000)                                                // Put x in range 1 <= x < 10.
    {
     while (x < 1.000000)
     {
      x *= 10.000;
      ie--;
     }

    }

 while (x >= 10.0000)
 {
  x = x / 10.0000;
  ie++;
 }

 ndig += ie;                                                      // In float format, number of digits is related to the size.
                                                                  // round x is between 1 and 10 and ndig will be printed to the right of decimal point.
 for (y = i = 1; i < ndig; i ++)
     {
      y = y / 10.0000;
     }

 x += y / 2.0000;

 if (x >= 10.0000)
    {
     x = 1.0000;
     ie++;
    }

 if (ie < 0)
    {
     *str++ = '0';
     *str++ = '.';
     if (ndig < 0)
        {
         ie = ie - ndig;
        }
     for (i = -1; i > ie; i --)
         {
          *str++ = '0';
         }
    }

 for (i = 0; i < ndig; i ++)
     {
      k = x;
      *str++ = k + '0';
      if (i == ie)
         {
          *str++ = '.';
         }
      x -= (y = k);
      x *= 10.0000;
     }

 *str = '\0';
}

void float2ascii_v2 (float x, unsigned char *str, char precision) // Converts a floating point number to an ASCII string.
{                                                                 // Version limited at 5 decimals maximum.
 int ie, i, k, ndig;                                              // x is stored into str, which should be at least 30 chars long.
 double y;

 if (precision >= 5)
    {
     precision = 5;
    }
    else
       {
        precision++;
       }

 ndig = precision;

 ie = 0;

 if (x < 0.00000)                                                 // If x is negative, write minus and reverse.
    {
     *str++ = '-';
     *str++ = ' ';
     x = -x;
    }

 if (x > 0.000000)                                                // Put x in range 1 <= x < 10.
    {
     while (x < 1.000000)
     {
      x *= 10.000;
      ie--;
     }

    }

 while (x >= 10.0000)
 {
  x = x / 10.0000;
  ie++;
 }

 ndig += ie;                                                      // In float format, number of digits is related to the size.
                                                                  // round x is between 1 and 10 and ndig will be printed to the right of decimal point.
 for (y = i = 1; i < ndig; i ++)
     {
      y = y / 10.0000;
     }

 x += y / 2.0000;

 if (x >= 10.0000)
    {
     x = 1.0000;
     ie++;
    }

 if (ie < 0)
    {
     *str++ = '0';
     *str++ = '.';
     if (ndig < 0)
        {
         ie = ie - ndig;
        }
     for (i = -1; i > ie; i --)
         {
          *str++ = '0';
         }
    }

 for (i = 0; i < ndig; i ++)
     {
      k = x;
      *str++ = k + '0';
      if (i == ie)
         {
          *str++ = '.';
         }
      x -= (y = k);
      x *= 10.0000;
     }

 *str = '\0';
}

void LCD_Init_v4 (void)
{
 Delay_ms(15);                                                    // LCD power ON initialization time >= 15 mS.
 LCD_Write(CMD, 0x30);                                            // 4 datas bits > Initialization of LCD with nibble method (4 datas bits).
 LCD_Write(CMD, 0x02);                                            // 4 datas bits > Initialization of LCD with nibble method (4 datas bits).
 LCD_Write(CMD, 0x28);                                            // 4 datas bits > 2 lines display, 5  8 dot character font.
 LCD_Write(CMD, 0x0C);                                            // 4 datas bits > Display ON. Cursor OFF.
 LCD_Write(CMD, 0x06);                                            // 4 datas bits > Auto increment cursor.
 LCD_Write(CMD, 0x01);                                            // 4 datas bits > Clear display.
 Delay_ms(1);                                                     // Ajustable ... (indispensable, sinon affichage erratique  la mise sous tension)
}

/******************************************************************************/

void main()
{
 PORTB = 0;                                                 // Initialisation du PORT B  0.
 PORTC = 0;                                                 // Initialisation du PORT C  0.
 PORTD = 0;                                                 // Initialisation du PORT D  0.

 ANSEL = 0b00001000;                                        // b3 = 1 (ANS3) : sets pin 5 (AN3) as analog input.
 ANSELH = 0b00000000;

 TRISB = 0b00000000;                                        // PORT B : b0  b7 configurs en sortie.
 TRISC = 0b00000000;                                        // PORT C : b0  b7 configurs en sortie.
 TRISD = 0b00000000;                                        // PORT D : b0  b7 configurs en sortie.

 C1ON_bit = 0;                                              // CMC1CON register > b7 > C1ON bit = 0 > Disable comparator 1.
 C2ON_bit = 0;                                              // CMC2CON register > b7 > C1ON bit = 0 > Disable comparator 2.

 SCS_bit = 0;                                               // OSCCON register > b0 > SCS bit = 0 > External oscillator (quartz) is used as a clock source.

 LCD_Init_v4 (void);

 pointeur_de_char = &texte[0];                              // pointeur_de_char pointe sur le premier lment du tableau "texte", soit texte[0].
                                                            // Autrement dit, pointeur_de_char contient l'adresse (&) de texte[0].
 do
  {
   int adc;
   float millivolts, temp;

   char txt[13];

   ADC_initialization();

   TRISD = 0x00;

    while(1)
    {
     adc = (ADC_read(3));                                   // Reads analog values.
     millivolts = adc * 4.888;                              // Converts analog values to millivolts (4.888 = 5000 mV / 1023). 5000 mV = 5V = VCC of MCU.
     temp = millivolts / 10.0;                              // With LM35, each 10 mV output correspond to 1C.

     StrConstRamCpy(pointeur_de_char, "Temperature ...  "); // OK.
     LCD_Write_String_At(0, 0, pointeur_de_char);           // OK. Voir dans void LCD_Write_String(char *msg), le while(*(msg + k) > 0).

     //float2ascii_v1(temp, &txt[0], 2);
     float2ascii_v2(temp, &txt[0], 2);

     //LCD_Write_String_At(1, 0, &txt[0]);                  // Write txt in 2nd row, starting at 1st digit.
     //LCD_Write_String_At(1, 13, " C");                   // Write " C" in 2nd row, starting at 14th digit.
     Lcd_Out(2, 0, &txt[0]);
     Lcd_Out(2, 13, " C");

     portc.b2 = 1;
     delay_ms(500);
     portc.b2 = 0;
     delay_ms(2500);
    }
  }
 while(1);
}
